#ifndef INCLUDE_H
#define INCLUDE_H

#ifdef _WIN32
	#define _USE_MATH_DEFINES
	#define WIN32_EXTRA_LEAN
	#define _CRT_SECURE_NO_DEPRECATE
	#define NOMINMAX
	#include <windows.h>
#endif


//std
#include <cstdio>
#include <cstring>
#include <vector>
#include <cmath>

using namespace std;

#include "vector.h"
#include "matrix.h"

float abs32(float val);

#define MY_PI 3.14159265359f
#define MY_HALF_PI 1.5707963268f

#pragma pack(1)
typedef struct bmfh
{
	short	type;
	int	size;
	short	reserved1;
	short	reserved2;
	int	offset;
} bmfh_t;
#pragma pack(8)

#pragma pack(1)
typedef struct bmih
{
	int	headerSize;
	long	width;
	long	height;
	short	planes;
	short	bpp;
	int	compression;
	int	imageSize;
	long	xres;
	long	yres;
	int	clrused;
	int	clrimportant;
} bmih_t;
#pragma pack(8)


#pragma pack(1)
typedef struct
{
	char type[2];
	int file_size;
	int reserved;
	int offset;
} bmpheader_t;
#pragma pack(8)

#pragma pack(1)
typedef struct
{
	int size;
	int width;
	int height;
	short planes;
	short bpp;
	int compression;
	int image_size;
	int xres;
	int yres;
	int clr_used;
	int clr_important;
} dib_t;
#pragma pack(8)

#pragma pack(1)
typedef struct
{
	bmpheader_t	header;
	dib_t		dib;
} bitmap_t;
#pragma pack(8)


// Lerp between A and B where time is within [0,1]
inline int lerp(int a, int b, float time)
{
	return a * (1 - time) + b * time;
}


typedef struct
{
	int *data;
	int width;
	int height;
} texinfo_t;


#define MAX(x,y) (x) > (y) ? (x) : (y)
#define MIN(x,y) (x) < (y) ? (x) : (y)

void draw_line(int *pixels, int width, int height, int x1, int y1, int x2, int y2, int color);
void flood_fill(int *pixels, int width, int height, int x, int y, int old_color, int new_color);
void flood_triangle(int *pixels, int width, int height, int x1, int y1, int x2, int y2, int x3, int y3, int color);
double fsin(double rad);
double fcos(double rad);

void barycentric_triangle_tex(int *pixels, float *zbuffer, int width, int height, texinfo_t *texture, int x1, int y1, float z1, float w1, int c1, int x2, int y2, float z2, float w2, int c2, int x3, int y3, float z3, float w3, int c3,
	float u1, float v1, float u2, float v2, float u3, float v3);
extern float zbuffer[1920 * 1080];
void barycentric_triangle(int *pixels, int width, int height, int x1, int y1, int z1, int c1, int x2, int y2, int z2, int c2, int x3, int y3, int z3, int c3);



inline void draw_pixel(int *pixels, float *zbuffer, int width, int height, int x, int y, int z, unsigned int color)
{
	if (width < 0 || height < 0)
		return;
	if (x >= width || y >= height)
		return;

	if (x < 0 || y < 0)
		return;


	if (zbuffer[x + y * width] > z)
	{
		pixels[x + ((height - 1 - y) * width)] = color;
		//pixels[x + (y * width)] = color;
		zbuffer[x + y * width] = z;
	}
}


inline int det(int ax, int ay, int bx, int by)
{
	return ax * by - bx *  ay;
}


#endif
