
#include "include.h"
#include <io.h>
#include <fcntl.h>

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

#define TICK_TIMER 1

int render(int *data, int width, int height, int time);

texinfo_t texinfo;


byte *tga_24to32(int width, int height, byte *pBits, bool bgr)
{
	int lImageSize = width * height * 4;
	byte *pNewBits = new byte[lImageSize * sizeof(byte)];

	for (int i = 0, j = 0; i < lImageSize; i += 4)
	{
		if (bgr)
		{
			pNewBits[i + 2] = pBits[j++];
			pNewBits[i + 1] = pBits[j++];
			pNewBits[i + 0] = pBits[j++];
		}
		else
		{
			pNewBits[i + 0] = pBits[j++];
			pNewBits[i + 1] = pBits[j++];
			pNewBits[i + 2] = pBits[j++];
		}
		pNewBits[i + 3] = 0;
	}
	return pNewBits;
}



void read_bitmap(char *filename, int &width, int &height, int **data)
{
	FILE *file;
	bitmap_t	bitmap;

	memset(&bitmap, 0, sizeof(bitmap_t));

	file = fopen(filename, "rb");
	if (file == NULL)
	{
		perror("Unable to write file");
		return;
	}

	fread(&bitmap, 1, sizeof(bitmap_t), file);
	width = bitmap.dib.width;
	height = bitmap.dib.height;
	fread((void *)*data, 1, width * height * 4, file);
	fclose(file);


	*data = (int *)tga_24to32(width, height, (byte *)*data, false);
}





int index_buffer[] = { 0,1,2,3,4,5,6,7,8,9,10, 11,12,13,14,15,16,17,18,19,20 ,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35};

int color[] = {
	RGB(255,0,0),
	RGB(0,255,0),
	RGB(0,0,255),
	RGB(255,0,255),
	RGB(0,255,255),
	RGB(255,255,255),
	RGB(255,0,0),
	RGB(0,255,0),
	RGB(0,0,255),
};

vec4 vert[36] =
{
	// Front face
	vec4(-0.5f, 0.5f, 0.5f, 1.0f), //3
	vec4(0.5f, -0.5f, 0.5f, 1.0f), //2
	vec4(0.5f, 0.5f, 0.5f, 1.0f),  //1

	vec4(-0.5f, -0.5f, 0.5f, 1.0f), //4
	vec4(0.5f, -0.5f, 0.5f, 1.0f),  //2
	vec4(-0.5f, 0.5f, 0.5f, 1.0f), //3

	// Back face
	vec4(0.5f, 0.5f, -0.5f, 1.0f),		//3
	vec4(-0.5f, -0.5f, -0.5f, 1.0f),	//2
	vec4(-0.5f, 0.5f, -0.5f, 1.0f),		//1

	vec4(0.5f, -0.5f, -0.5f, 1.0f),		//4
	vec4(-0.5f, -0.5f, -0.5f, 1.0f),	//2
	vec4(0.5f, 0.5f, -0.5f, 1.0f),		//3

	// Left face
	vec4(-0.5f, 0.5f, -0.5f, 1.0f),		//3
	vec4(-0.5f, -0.5f, 0.5f, 1.0f),		//2
	vec4(-0.5f, 0.5f, 0.5f, 1.0f),		//1

	vec4(-0.5f, -0.5f, -0.5f, 1.0f),	//4
	vec4(-0.5f, -0.5f, 0.5f, 1.0f),		//2
	vec4(-0.5f, 0.5f, -0.5f, 1.0f),		//3

	// Right face
	vec4(0.5f, 0.5f, 0.5f, 1.0f),		//3
	vec4(0.5f, -0.5f, -0.5f, 1.0f),		//2
	vec4(0.5f, 0.5f, -0.5f, 1.0f),		//1

	vec4(0.5f, -0.5f, 0.5f, 1.0f),		//4
	vec4(0.5f, -0.5f, -0.5f, 1.0f),		//2
	vec4(0.5f, 0.5f, 0.5f, 1.0f),		//3

	// Top face
	vec4(-0.5f, 0.5f, 0.5f, 1.0f),		//3
	vec4(0.5f, 0.5f, -0.5f, 1.0f),		//2
	vec4(-0.5f, 0.5f, -0.5f, 1.0f),		//1

	vec4(0.5f, 0.5f, 0.5f, 1.0f),		//4
	vec4(0.5f, 0.5f, -0.5f, 1.0f),		//2
	vec4(-0.5f, 0.5f, 0.5f, 1.0f),		//3

	// Bottom face
	vec4(0.5f, -0.5f, 0.5f, 1.0f),		//3
	vec4(-0.5f, -0.5f, -0.5f, 1.0f),	//2
	vec4(0.5f, -0.5f, -0.5f, 1.0f),		//1

	vec4(-0.5f, -0.5f, 0.5f, 1.0f),		//4
	vec4(-0.5f, -0.5f, -0.5f, 1.0f),	//2
	vec4(0.5f, -0.5f, 0.5f, 1.0f)		//3
};

vec2 texcord[36] =
{
	// Front face
	vec2(0.0f, 0.0f),			//2
	vec2(0.0f, -1.0f),			//1
	vec2(-1.0f, -1.0f),			//3

	vec2(0.0f, 0.0f),			//2
	vec2(-1.0f, -1.0f),			//3
	vec2(-1.0f, 0.0f),			//4


	// Back face
	vec2(0.0f, 0.0f),			//2
	vec2(0.0f, 1.0f),			//1
	vec2(1.0f, 1.0f),			//3

	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 1.0f),			//3
	vec2(1.0f, 0.0f),			//4

	// Left face
	vec2(0.0f, -1.0f),			//1
	vec2(0.0f, 0.0f),			//2
	vec2(-1.0f, -1.0f),			//3

	vec2(0.0f, 0.0f),			//2
	vec2(-1.0f, 0.0f),			//4
	vec2(-1.0f, -1.0f),			//3

	// Right face
	vec2(0.0f, 1.0f),			//1
	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 1.0f),			//3

	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 0.0f),			//4
	vec2(1.0f, 1.0f),			//3

	// Top face
	vec2(0.0f, 1.0f),			//1
	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 1.0f),			//3

	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 0.0f),			//4
	vec2(1.0f, 1.0f),			//3

	// Bottom face
	vec2(0.0f, 1.0f),			//1
	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 1.0f),			//3

	vec2(0.0f, 0.0f),			//2
	vec2(1.0f, 0.0f),			//4
	vec2(1.0f, 1.0f),			//3
};


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, INT iCmdShow)
{
	HWND		hwnd;
	HACCEL		hAccel;
	MSG		msg;
	WNDCLASS	wndclass;
	char		szAppName[] = TEXT("WinMain");

	wndclass.style		= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndclass.lpfnWndProc	= WndProc;
	wndclass.cbClsExtra	= 0;
	wndclass.cbWndExtra	= 0;
	wndclass.hInstance	= hInstance;
	wndclass.hIcon		= LoadIcon(hInstance, "altEngine");
	wndclass.hCursor	= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= NULL;
	wndclass.lpszMenuName	= NULL;
	wndclass.lpszClassName	= szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Unable to register window class."), szAppName, MB_ICONERROR);
		return 0;
	}
	hwnd = CreateWindow(	szAppName,			// window class name
				TEXT("Software Render"),		// window caption
				WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,		// window style
				CW_USEDEFAULT,			// initial x position
				CW_USEDEFAULT,			// initial y position
				512,			// initial x size
				512,			// initial y size
				NULL,				// parent window handle
				NULL,				// window menu handle
				hInstance,			// program instance handle
				NULL);				// creation parameters

	hAccel = LoadAccelerators(hInstance, szAppName);
	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);


	texinfo.data = new int[512 * 512];
	texinfo.width = 512;
	texinfo.height = 512;



	int checksize = 8;
	for (int y = 0; y < 512; y++)
	{
		for (int x = 0; x < 512; x++)
		{
			int total = floor(x / 512.0f * checksize) +
				floor(y / 512.0f * checksize);
			bool isEven = total % 2 == 0;

			if (isEven)
				texinfo.data[x + y * 512] = ~0;
			else
				texinfo.data[x + y * 512] = 0;

		}
	}

	read_bitmap("tex.bmp", texinfo.width, texinfo.height, &texinfo.data);
     
	while (TRUE)
	{
		if ( GetMessage(&msg, NULL, 0, 0) > 0 )
		{
			if (!TranslateAccelerator(hwnd, hAccel, &msg))
			{
				TranslateMessage (&msg) ;
				DispatchMessage (&msg) ;
			}
		}
		else
		{
			break;
		}
	}

	if (!UnregisterClass(szAppName, hInstance))
	{
		MessageBox(NULL, TEXT("Unable to unregister class."), szAppName, MB_ICONERROR);
		return 1;
	}

	return (int)msg.wParam;
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HDC	hdc;
	static HDC	hdcMem;
	static POINT	center;
	static HBITMAP	hBitmap;
	static HGDIOBJ hObject;
	static BITMAP	bitmap;
	static int	width, height;
	static int	*data;
	static int	time = 0;


	switch (message)
	{
	case WM_CREATE:
		hdc = GetDC(hwnd);
		hdcMem = CreateCompatibleDC(hdc);
		hBitmap = CreateCompatibleBitmap(hdc, width, height);
		hObject = SelectObject(hdcMem, hBitmap);
		SetTimer( hwnd, TICK_TIMER, 16, NULL );
		return 0;

	case WM_TIMER:
		switch(wParam)
		{
		case TICK_TIMER:
			time++;
//			GetBitmapBits(hBitmap, width * height * sizeof(int), data);
			memset(data, 0, width * height * sizeof(int));
			render(data, width, height, time);
			SetBitmapBits(hBitmap, width * height * sizeof(int), data);
			BitBlt(hdc, 0, 0, width, height, hdcMem, 0, 0, SRCCOPY);
			break;
		}
		return 0;

	case WM_MOUSEMOVE:
		{

		}
		return 0;
	case WM_SIZE:
		{
			SelectObject(hdcMem, hObject);
			DeleteObject(hBitmap);
			DeleteDC(hdcMem);

			width	= LOWORD(lParam);
			height	= HIWORD(lParam);
			free((void *)data);
			data = (int *)malloc(width * height * sizeof(int));
			center.x = width / 2;
			center.y = height / 2;

			hdcMem = CreateCompatibleDC(hdc);
			hBitmap = CreateCompatibleBitmap(hdc, width, height);
			hObject = SelectObject(hdcMem, hBitmap);
		}
		return 0;

	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc (hwnd, message, wParam, lParam);
}

void rotate_vector(float rad, vec3 &vec, vec3 &axis)
{
	matrix3	m;
	float	sinVal, cosVal, minusVal;

	sinVal = (float)fsin(rad);
	cosVal = (float)fcos(rad);
	minusVal = 1.0f - cosVal;

	m.m[0] = cosVal + minusVal * axis.x * axis.x;
	m.m[1] = minusVal * axis.x * axis.y - sinVal * axis.z;
	m.m[2] = minusVal * axis.x * axis.z + sinVal * axis.y;

	m.m[3] = minusVal * axis.y * axis.x + sinVal * axis.z;
	m.m[4] = cosVal + minusVal * axis.y * axis.y;
	m.m[5] = minusVal * axis.y * axis.z - sinVal * axis.x;

	m.m[6] = minusVal * axis.z * axis.x - sinVal * axis.y;
	m.m[7] = minusVal * axis.z * axis.y + sinVal * axis.x;
	m.m[8] = cosVal + minusVal * axis.z * axis.z;

	vec = m * vec;
}

int render(int *data, int width, int height, int time)
{
	matrix4 transformation;
	matrix4 projection;
	matrix4 mvp;
	vec4 box2d[36];

	vec3 pos(1.0f, 1.0f, -0.5f);


	vec3 forward = vec3(0.0f, 0.0, 1.0f);
	vec3 up = vec3(0.0f, 1.0f, 0.0f);
	vec3 vup = vec3(0.0f, 1.0f, 0.0f);
	vec3 right;

	// Left / Right
	rotate_vector(time * 0.01f, forward, vup );
	rotate_vector(time * 0.01f, up, vup);
	forward.normalize();
	up.normalize();

	// Up / Down
	right = vec3::crossproduct(up, forward);
	right.normalize();

	rotate_vector(time * 0.01f, forward, right);
	rotate_vector(time * 0.01f, up, right);
	forward.normalize();
	up.normalize();


	transformation.m[0] = right.x;
	transformation.m[1] = right.y;
	transformation.m[2] = right.z;
	transformation.m[3] = 0.0f;

	transformation.m[4] = up.x;
	transformation.m[5] = up.y;
	transformation.m[6] = up.z;
	transformation.m[7] = 0.0f;

	transformation.m[8] = forward.x;
	transformation.m[9] = forward.y;
	transformation.m[10] = forward.z;
	transformation.m[11] = 0.0f;

	transformation.m[12] = 0.0f;
	transformation.m[13] = 0.0f;
	transformation.m[14] = 0.0f;
	transformation.m[15] = 1.0f;

	// typical opengl style projection
	projection.perspective(90.0f, 1.0f, 1.0, 2001.0f, false);

	mvp = transformation;

	//rotate about origin
	for(int i = 0; i < 36; i++)
	{
		box2d[i] = mvp * vert[i];
	}

	transformation.m[0] = 1.0f;
	transformation.m[1] = 0.0f;
	transformation.m[2] = 0.0f;
	transformation.m[3] = 0.0f;

	transformation.m[4] = 0.0f;
	transformation.m[5] = 1.0f;
	transformation.m[6] = 0.0f;
	transformation.m[7] = 0.0f;

	transformation.m[8] = 0.0f;
	transformation.m[9] = 0.0f;
	transformation.m[10] = 1.0f;
	transformation.m[11] = 0.0f;

	transformation.m[12] = pos.x;
	transformation.m[13] = pos.y;
	transformation.m[14] = pos.z;
	transformation.m[15] = 1.0f;

	// make things bigger
	transformation = transformation * (height / 2.0f);

	mvp = transformation * projection;

	//translate a little, scale and project
	for(int i = 0; i < 36; i++)
	{
		box2d[i] = mvp * box2d[i];
	}


	for (int i = 0; i < 36; i+=3)
	{
			barycentric_triangle_tex(data, zbuffer, width, height, &texinfo,
				box2d[index_buffer[i]].x, box2d[index_buffer[i]].y, box2d[index_buffer[i]].z, box2d[index_buffer[i]].w, RGB(255, 0, 0),
				box2d[index_buffer[i+1]].x, box2d[index_buffer[i+1]].y, box2d[index_buffer[i + 1]].z, box2d[index_buffer[i + 1]].w, RGB(255, 0, 0),
				box2d[index_buffer[i+2]].x, box2d[index_buffer[i+2]].y, box2d[index_buffer[i + 2]].z, box2d[index_buffer[i + 2]].w, RGB(255, 0, 0),
				texcord[index_buffer[i]].x, texcord[index_buffer[i]].y,
				texcord[index_buffer[i+1]].x, texcord[index_buffer[i+1]].y,
				texcord[index_buffer[i+2]].x, texcord[index_buffer[i+2]].y);
	}
	memset(zbuffer, 0, width*height * 4);
	return 0;
}

